import pytest
from math import comb

from kappybara.algebra import Expression
from kappybara.pattern import Pattern
from kappybara.rule import KappaRule, KappaRuleUnimolecular, KappaRuleBimolecular
from kappybara.system import System


@pytest.mark.parametrize(
    "test_case",
    [
        ("A(), B()", 10, KappaRule, "B(), A()", 100),
        ("A(), B()", 10, KappaRuleUnimolecular, "B(), A()", 0),
        ("A(), B()", 10, KappaRuleBimolecular, "B(), A()", 100),
        ("A()", 10, KappaRule, "A(), A()", 100),  # No automorphism checks currently
        ("A(a[1]), B(b[1]), C()", 10, KappaRule, "A(a[1]), B(b[1]), C()", 100),
        ("A(a[1]), B(b[1]), C()", 10, KappaRule, "A(), B(), C()", 1000),
        ("A(a[1]), B(b[1])", 10, KappaRuleUnimolecular, "B(), A()", 10),
        ("A(a[1]), B(b[1])", 10, KappaRuleBimolecular, "B(), A()", 90),
        (
            "A(a1[1]), B(b1[1], b2[2]), B(b1[2], b2[3]) A(a2[3])",
            10,
            KappaRuleUnimolecular,
            "B(), A()",
            40,
        ),
    ],
)
def test_rule_n_embeddings_at_system_initialiation(test_case):
    (mixture_pattern_str, n_copies, rule_class, rule_pattern_str, n_embeddings) = (
        test_case
    )
    rule_pattern = Pattern.from_kappa(rule_pattern_str)
    rule = rule_class(rule_pattern, rule_pattern, Expression.from_kappa("1.0"))
    system = System.from_kappa({mixture_pattern_str: n_copies}, [rule.kappa_str])
    assert system.rules["r0"].n_embeddings(system.mixture) == n_embeddings


def test_simple_rule_application():
    """Test selection/application of a simple KappaRule in a mixture."""
    n_copies = 10
    rule_left_str = "A(a[1]), B(b[1])"
    rule = KappaRule(
        Pattern.from_kappa(rule_left_str),
        Pattern.from_kappa("A(a[.]), B(b[.])"),
        Expression.from_kappa("1.0"),
    )
    observables = [f"|{rule_left_str}|", "|A(a[.])|", "|B(b[#])|"]
    system = System.from_kappa(
        {"A(a[1]), B(b[1])": n_copies}, [rule.kappa_str], observables
    )
    rule = system.rules["r0"]

    assert rule.n_embeddings(system.mixture) == n_copies
    assert system["o0"] == n_copies
    assert system["o1"] == 0

    for i in range(1, n_copies + 1):
        update = rule.select(system.mixture)
        assert len(update.edges_to_remove) == 1

        system.mixture.apply_update(update)
        assert system["o0"] == n_copies - i
        assert system["o1"] == i
        assert system["o2"] == n_copies


def test_edge_creating_rule_application():
    """Test selection/application of a KappaRule which creates a new edge in a mixture."""
    n_copies = 4
    rule_right_str = "A(a[1]), B(b[1])"
    rule = KappaRule(
        Pattern.from_kappa("A(a[.]), B(b[.])"),
        Pattern.from_kappa(rule_right_str),
        Expression.from_kappa("1.0"),
    )
    observables = [f"|{rule_right_str}|"]
    system = System.from_kappa(
        {"A(a[.]), B(b[.])": n_copies}, [rule.kappa_str], observables
    )
    rule = system.rules["r0"]

    assert rule.n_embeddings(system.mixture) == n_copies * n_copies
    assert system["o0"] == 0
    for _ in range(1, n_copies + 1):
        update = rule.select(system.mixture)
        assert len(update.edges_to_add) == 1
        system.mixture.apply_update(update)
    assert system["o0"] == n_copies


def test_rule_application():
    """
    Test selection/application of a slightly more involved KappaRule in a mixture.

    TODO: Support for empty slots (".") in pattern strings when instantiating from Lark.
    TODO: Supporting agent interfaces so we know a master list of sites for agents to be able to initialize defaults.
    """
    n_copies = 100
    rule_left_str = "A(a[1]), B(b[1], x[3]), C(c[2]{p}), D(d[2]{p}, x[3])"
    rule = KappaRule(
        Pattern.from_kappa(rule_left_str),
        Pattern.from_kappa("A(a[1]), B(b[.], x[3]), C(c[1]{u}), D(d[.]{p}, x[3])"),
        Expression.from_kappa("1.0"),
    )
    observables = [f"|{rule_left_str}|", "|A(a[1]), C(c[1])|", "|B(b[_])|", "|C(c{u})|"]
    system = System.from_kappa(
        {"A(a[1]), B(b[1], x[3]), C(c[2]{p}), D(d[2]{p}, x[3])": n_copies},
        [rule.kappa_str],
        observables,
    )
    rule = system.rules["r0"]

    assert rule.n_embeddings(system.mixture) == n_copies
    assert system["o0"] == n_copies
    assert system["o1"] == 0

    for i in range(1, n_copies + 1):
        update = rule.select(system.mixture)
        assert len(update.edges_to_remove) == 2
        assert len(update.edges_to_add) == 1
        assert len(update.agents_changed) == 1

        system.mixture.apply_update(update)
        assert system["o0"] == n_copies - i
        assert system["o1"] == i
        assert system["o2"] == n_copies - i
        assert system["o3"] == i


@pytest.mark.parametrize("n_copies", [50])
def test_simple_unimolecular_rule_application(n_copies):
    """Test selection/application of a simple unimolecular KappaRule in a mixture."""
    system = System.from_kappa(
        {"A(a[1]{u}), B(b[1]{u})": n_copies},
        [
            "A(a{u}), B(b{u}) -> A(a{p}), B(b{p}) @ 0.0 {1.0}",
            "A(a[1]), B(b[1]) -> A(a[.]), B(b[.]) @ 1.0",
        ],
        ["|A(a[1]{u}), B(b[1]{u})|"],
    )

    rule1 = system.rules["r0"]
    rule2 = system.rules["r1"]
    assert isinstance(rule1, KappaRuleUnimolecular)
    assert isinstance(rule2, KappaRule)

    assert system["o0"] == n_copies
    n_rule1_applications = n_copies // 2
    n_rule2_applications = n_copies // 2

    for i in range(1, n_rule2_applications + 1):
        update = rule2.select(system.mixture)
        system.mixture.apply_update(update)
        assert system["o0"] == n_copies - i
        assert len(system.mixture.components) == n_copies + i
        assert rule1.n_embeddings(system.mixture) == n_copies - i

    for i in range(1, n_rule1_applications + 1):
        rule1.n_embeddings(
            system.mixture
        )  # Uni/bimolecular rules use this to weight choices
        update = rule1.select(system.mixture)
        system.mixture.apply_update(update)
        assert rule1.n_embeddings(system.mixture) == n_copies - n_rule2_applications - i
        assert system["o0"] == n_copies - n_rule2_applications - i


@pytest.mark.parametrize("n_copies", [50])
def test_simple_bimolecular_rule_application(n_copies):
    """Test selection/application of a simple bimolecular KappaRule in a mixture."""
    system = System.from_kappa(
        {"A(a[.]{u})": n_copies},
        ["A(a{u}), A(a{u}) -> A(a{p}), B(a{p}) @ 1.0 {0.0}"],
        ["|B(a{p})|"],
    )

    rule1 = system.rules["r0"]
    assert isinstance(rule1, KappaRuleBimolecular)

    n_rule1_applications = n_copies // 2
    for i in range(1, n_rule1_applications + 1):
        rule1.n_embeddings(
            system.mixture
        )  # Uni/bimolecular rules use this to weight choices
        update = rule1.select(system.mixture)
        system.mixture.apply_update(update)
        assert rule1.n_embeddings(system.mixture) == 2 * comb(n_copies - 2 * i, 2)
        assert system["o0"] == i



import pytest
import shutil
import itertools
from collections import defaultdict

from kappybara.rule import AVOGADRO, DIFFUSION_RATE, kinetic_to_stochastic_on_rate
from kappybara.system import System
from kappybara.examples import heterodimerization_system


def test_basic_system():
    system = System.from_kappa(
        {"A(a[.], b[.])": 100},
        rules=[
            "A(a[.]), A(a[.]) <-> A(a[1]), A(a[1]) @ 1.0 {2.0}, 1.0",
            "A(b[.]), A(b[.]) <-> A(b[1]), A(b[1]) @ 1.5, 1.0",
        ],
        observables=[
            "|A(a[.])|",
            "|A(b[1]), A(b[1])|",
            "|A(a[1], b[.]), A(a[1], b[_])|",
        ],
    )

    counts = defaultdict(list)
    for _ in range(1000):
        system.update()
        for obs_name in system.observables:
            counts[obs_name].append(system[obs_name])


def test_system_from_kappa():
    system = System.from_ka(
        """
    %def: "maxConsecutiveClash" "20"
    %def: "seed" "365457"

    // constants
    %var: 'x'     0.03
    %var: 'k_on'  'x' * 10
    %var: 'g_on'  'k_on' / 100

    %var: 'n' 3 * 100

    %init: 'n' A(a[1]{p}), B(b[1]{u})

    %obs: 'A_total'   |A()|
    %obs: 'A_u'       |A(a{u})|
    %obs: 'B_u'       |B(b{u})|
    %obs: 'A_p'       |A(a{p})|
    %obs: 'pairs'     |A(a[1]), B(b[1])|

    A(a{p}), B(b[_]) -> A(a{u}), B() @ 'g_on'
    """
    )
    n = system["n"]
    assert n == 300
    assert system["g_on"] == 0.003
    assert system["A_total"] == n

    for i in range(1, n):
        system.update()
        assert system["A_total"] == n
        assert system["A_u"] == i
        assert system["B_u"] == n
        assert system["A_p"] == n - i
        assert system["pairs"] == n


@pytest.mark.parametrize(
    "k_on, expected, via_kasim",
    [
        (k_on, expected, vk)
        for vk in ((False, True) if shutil.which("KaSim") else (False,))
        for k_on, expected in [(2.5e8, 65), (2.0e9, 331)]
    ],
)
def test_heterodimerization(k_on, expected, via_kasim):
    system = heterodimerization_system(k_on)
    n_heterodimers = []
    while system.time < 2:
        if via_kasim:
            system.update_via_kasim(0.1)
        else:
            system.update()
        if system.time > 1:
            n_heterodimers.append(system["o0"])

    measured = sum(n_heterodimers) / len(n_heterodimers)
    assert abs(measured - expected) < expected / 5


@pytest.mark.parametrize(
    "kd, a_init, b_init",
    itertools.product([10**-9], [2000], [2000, 3500]),
)
def test_equilibrium_matches_kd(kd, a_init, b_init):
    """
    Check that the input Kd matches what's observed empirically post-equilibrium
    within a relative margin of error.
    """
    volume = 10**-13
    on_rate = kinetic_to_stochastic_on_rate(volume=volume)
    kd = 10**-9
    off_rate = DIFFUSION_RATE * kd
    system = System.from_ka(
        f"""
        %init: {a_init} A(x[.])
        %init: {b_init} B(x[.])
        %obs: 'A' |A(x[.])|
        %obs: 'B' |B(x[.])|
        %obs: 'AB' |B(x[_])|
        A(x[.]), B(x[.]) <-> A(x[1]), B(x[1]) @ {on_rate}, {off_rate}
        """
    )

    empirical_kds = []
    while system.time < 2:
        system.update()
        a_conc_eq = system["A"] / AVOGADRO / volume
        b_conc_eq = system["B"] / AVOGADRO / volume
        ab_conc_eq = system["AB"] / AVOGADRO / volume
        empirical_kds.append(a_conc_eq * b_conc_eq / ab_conc_eq)
    i = int(len(empirical_kds) * 0.5)  # an index post-equilibrium
    empirical_kd = sum(empirical_kds[i:]) / len(empirical_kds[i:])
    assert abs((empirical_kd - kd) / kd) < 0.1


def test_system_manipulation():
    system = System.from_ka(
        """
        %init: 10 A(x[.])
        %init: 10 B(x[.])
        %init: 1 C()

        %obs: 'A' |A(x[.])|
        %obs: 'B' |B(x[.])|
        %obs: 'AB' |A(x[1]), B(x[1])|

        %var: 'total_agents' 'A' + 'B' + (2 * 'AB')

        A(x[.]), B(x[.]) -> A(x[1]), B(x[1]) @ 1 {1}
        """
    )
    assert not system["AB"]
    system.update()
    assert system["AB"] == 1

    # Add a pattern
    system.mixture.instantiate("A(x[1]), B(x[1])")
    assert system["AB"] == 2
    total_agents_pre_removal = system["total_agents"]
    assert total_agents_pre_removal == 22

    # Remove a component
    component_to_remove = system.mixture.components[0]
    system.mixture.remove(component_to_remove)
    assert total_agents_pre_removal - system["total_agents"] == len(component_to_remove)

    # Add the component back
    system.mixture.add(component_to_remove)
    assert system["total_agents"] == total_agents_pre_removal

    # Set a new variable
    system["twiceA"] = "2 * 'A'"
    assert system["twiceA"] == 2 * system["A"]
    system.update()
    assert system["twiceA"] == 2 * system["A"]

    # Update an observable
    system["A"] = "-1"
    assert system["twiceA"] == -2

    # Set a new observable
    system["C"] = "|C()|"
    assert system["C"] == 1
    system.update()
    assert system["C"] == 1

    system.add_rule("new", "B() -> C() @ 1000")
    while system.reactivity:
        system.update()
    assert system["B"] == 0 and system["C"] == 12
    system.remove_rule("new")
    system.add_rule("newer", "C() -> B() @ 1000")
    while system.reactivity:
        assert system["B"] == 12 and system["C"] == 0



import random
from typing import Any, Optional, Iterable, Generic, TypeVar, Self
from collections import defaultdict
from collections.abc import Callable, Hashable


def rejection_sample(population: Iterable, excluded: Iterable, max_attempts: int = 100):
    population = list(population)
    if not population:
        raise ValueError("Sequence is empty")
    excluded_ids = set(id(x) for x in excluded)

    # Fast rejection sampling (O(1) average case for small exclusion sets)
    for _ in range(max_attempts):
        choice = random.choice(population)
        if id(choice) not in excluded_ids:
            return choice

    # Fallback to O(n) scan only if necessary (rare for small exclusion sets)
    valid_choices = [item for item in population if id(item) not in excluded_ids]
    if not valid_choices:
        raise ValueError("No valid elements to choose from")
    return random.choice(valid_choices)


class OrderedSet[T]:
    def __init__(self, items: Optional[Iterable[T]] = None):
        self.dict = dict() if items is None else dict.fromkeys(items)

    def __iter__(self):
        yield from self.dict

    def __len__(self):
        return len(self.dict)

    def add(self, item: Any) -> None:
        self.dict[item] = None

    def remove(self, item: Any) -> None:
        del self.dict[item]


class Counted:
    counter = 0

    def __init__(self):
        self.id = Counted.counter
        Counted.counter += 1

    def __hash__(self):
        return self.id

    def __eq__(self, other):
        return hash(self) == hash(other)


# --- Indexed sets ---

T = TypeVar("T")  # Member type of `IndexedSet`


class SetProperty:
    """
    This class should be initialized with a function or lambda which
    takes a set member as input, and returns a collection or iterable of
    values associated with that member that you want to index by.

    Example initialization:
    ```
    @dataclass
    class SportsTeam:
        name: str
        members: list[str]

    members_alt = SetProperty(lambda team: team.members) # If someone can belong to multiple teams (default)
    ```
    """

    def __init__(self, fn: Callable[[T], Iterable[Hashable]], is_unique=False):
        self.fn = fn
        self.is_unique = is_unique

    def __call__(self, item: T) -> Iterable[Hashable]:
        return self.fn(item)


class Property(SetProperty):
    """
    This class should be initialized with a function or lambda which
    takes a set member as input, and returns a single object which is
    the corresponding property value of a set member.

    An example of initializing a `Property`
    ```
    @dataclass
    class Fruit:
        color: str

    my_property = Property(lambda fruit: fruit.color) # Equivalent using a lambda function
    ```
    """

    def __call__(self, item: T) -> Iterable[Hashable]:
        return [self.fn(item)]


class IndexedSet(set[T], Generic[T]):
    """
    A subclass of the built-in `set`, with support for indexing
    by arbitrary properties of set members, as well as integer
    indexing to allow for random sampling.

    Credit https://stackoverflow.com/a/15993515 for the integer indexing logic.

    If you know for some property that you should only get a single
    set member back when using `lookup`, mark that property as unique
    when you create it.

    NOTE: although this class is indexable due to the implementation
    of `__getitem__`, member ordering is not stable across insertions
    and deletions.

    Example usage:
    ```
    @dataclass
    class SportsTeam:
        name: str
        jersey_color: str
        members: list[str]

    teams: IndexedSet[SportsTeam] = IndexedSet()
    teams.create_index("name", Property(lambda team: team.name, is_unique=True))
    teams.create_index("color", Property(lambda team: team.jersey_color))

    [...] # populate the set with teams

    teams.lookup("name", "Manchester") # Returns the team whose name is "Manchester"
    teams.lookup("color", "blue")    # Returns all teams with blue jerseys
    ```
    """

    properties: dict[str, SetProperty]
    indices: dict[str, defaultdict[Hashable, Self]]

    _item_to_pos: dict[T, int]
    _item_list: list[T]

    def __init__(self, iterable: Iterable[T] = []):
        iterable = list(iterable)
        super().__init__(iterable)
        self._item_list = iterable
        self._item_to_pos = {item: i for (i, item) in enumerate(iterable)}

        self.properties = {}
        self.indices = {}

    def add(self, item: T):
        if item in self:
            return
        super().add(item)

        # Update integer index
        self._item_list.append(item)
        self._item_to_pos[item] = len(self._item_list) - 1

        # Update property indices
        for prop_name in self.properties:
            prop = self.properties[prop_name]
            for val in prop(item):
                if prop.is_unique:
                    assert not self.indices[prop_name][val]
                self.indices[prop_name][val].add(item)

    def remove(self, item: T):
        assert item in self
        super().remove(item)

        # Update integer index
        pos = self._item_to_pos.pop(item)
        last_item = self._item_list.pop()
        if pos != len(self._item_list):
            self._item_list[pos] = last_item
            self._item_to_pos[last_item] = pos

        # Update property indices
        for prop_name in self.properties:
            prop = self.properties[prop_name]
            for val in prop(item):
                self.indices[prop_name][val].remove(item)
                # If the index entry is now empty, delete it
                if not self.indices[prop_name][val]:
                    del self.indices[prop_name][val]

    def lookup(self, name: str, value: Any) -> T | Iterable[T]:
        prop = self.properties[name]
        matches = self.indices[name][value]

        if prop.is_unique:
            assert len(matches) == 1
            return next(iter(matches))
        else:
            return matches

    def remove_by(self, prop_name: str, value: Any):
        """
        Remove all set members whose property `prop_name` matches or contains `value`.
        """
        if value not in self.indices[prop_name]:
            return
        matches = list(self.indices[prop_name][value])
        for match in matches:
            assert match in self
            self.remove(match)

    def create_index(self, name: str, prop: SetProperty):
        """
        Given a function which maps a set member to an `Any`-typed value, create
        a reverse-index mapping a property value to the set of the members of `self`
        with that property. This index is updated when adding new members or removing
        existing ones, but please note that if you mutate the internal state of an
        existing set member, this object will not reflect those updates unless you
        take the care to update the indices manually.

        NOTE: mutating set members outside of interface calls can invalidate indices.
        """
        assert name not in self.properties
        self.properties[name] = prop
        self.indices[name] = defaultdict(IndexedSet)

        for el in self:
            for val in prop(el):
                if prop.is_unique:
                    assert not self.indices[name][val]
                self.indices[name][val].add(el)

    def __getitem__(self, i):
        assert 0 <= i < len(self)
        return self._item_list[i]



from pathlib import Path
from dataclasses import dataclass
from typing import Optional
from lark import Lark, ParseTree, Tree, Visitor, Token, Transformer_NonRecursive

from kappybara.pattern import Site, Agent, Pattern, SiteType, Partner
from kappybara.rule import Rule, KappaRule, KappaRuleUnimolecular, KappaRuleBimolecular
from kappybara.algebra import Expression


class KappaParser:
    """Parser for Kappa language files and expressions.

    Note:
        Don't instantiate directly: use the global kappa_parser instance.
    """

    def __init__(self):
        """Initialize the Lark parser with Kappa grammar."""
        self._parser = Lark.open(
            str(Path(__file__).parent / "kappa.lark"),
            rel_to=__file__,
            parser="earley",
            # The basic lexer isn't required and isn't usually recommended
            lexer="dynamic",
            start="kappa_input",
            # Disabling these slightly improves speed
            propagate_positions=False,
            maybe_placeholders=False,
        )

    def parse(self, text: str) -> ParseTree:
        return self._parser.parse(text)

    def parse_file(self, filepath: str) -> ParseTree:
        with open(filepath, "r") as file:
            return self._parser.parse(file.read())


kappa_parser = KappaParser()


@dataclass
class SiteBuilder(Visitor):
    """Builds Site objects from Lark parse trees.

    Attributes:
        parsed_site_name: Name of the site being built.
        parsed_state: Internal state of the site.
        parsed_partner: Partner specification for the site.
    """

    parsed_site_name: str
    parsed_state: str
    parsed_partner: Partner

    def __init__(self, tree: ParseTree):
        super().__init__()

        self.parsed_agents: list["Agent"] = []

        assert tree.data == "site"
        self.visit(tree)

    # Visitor method for Lark
    def site_name(self, tree: ParseTree) -> None:
        self.parsed_site_name = str(tree.children[0])

    # Visitor method for Lark
    def state(self, tree: ParseTree) -> None:
        match tree.children[0]:
            case "#":
                self.parsed_state = "#"
            case str(state):
                self.parsed_state = str(state)
            case Tree(data="unspecified"):
                self.parsed_state = "?"
            case _:
                raise ValueError(
                    f"Unexpected internal state in site parse tree: {tree}"
                )

    # Visitor method for Lark
    def partner(self, tree: ParseTree) -> None:
        match tree.children:
            case ["#"]:
                self.parsed_partner = "#"
            case ["_"]:
                self.parsed_partner = "_"
            case ["."]:
                self.parsed_partner = "."
            case [Token("INT", x)]:
                self.parsed_partner = int(x)
            case [
                Tree(data="site_name", children=[site_name]),
                Tree(data="agent_name", children=[agent_name]),
            ]:
                self.parsed_partner = SiteType(str(site_name), str(agent_name))
            case [Tree(data="unspecified")]:
                self.parsed_partner = "?"
            case _:
                raise ValueError(f"Unexpected link state in site parse tree: {tree}")

    @property
    def object(self) -> Site:
        return Site(
            label=self.parsed_site_name,
            state=self.parsed_state,
            partner=self.parsed_partner,
        )


@dataclass
class AgentBuilder(Visitor):
    """Builds Agent objects from Lark parse trees.

    Attributes:
        parsed_type: Type name of the agent.
        parsed_interface: List of sites belonging to the agent.
    """

    parsed_type: str
    parsed_interface: list[Site]

    def __init__(self, tree: ParseTree):
        super().__init__()

        self.parsed_type = None
        self.parsed_interface: list[Site] = []

        assert tree.data == "agent"
        self.visit(tree)

    # Visitor method for Lark
    def agent_name(self, tree: ParseTree) -> None:
        self.parsed_type = str(tree.children[0])

    # Visitor method for Lark
    def site(self, tree: ParseTree) -> None:
        self.parsed_interface.append(SiteBuilder(tree).object)

    @property
    def object(self) -> Agent:
        agent = Agent(type=self.parsed_type, sites=self.parsed_interface)
        for site in agent:
            site.agent = agent
        return agent


@dataclass
class PatternBuilder(Visitor):
    """Builds Pattern objects from Lark parse trees.

    Attributes:
        parsed_agents: List of agents in the pattern.
    """

    parsed_agents: list[Agent]

    def __init__(self, tree: ParseTree):
        super().__init__()

        self.parsed_agents: list[Agent] = []

        assert tree.data == "pattern"
        self.visit(tree)

    # Visitor method for Lark
    def agent(self, tree: ParseTree) -> None:
        self.parsed_agents.append(AgentBuilder(tree).object)

    @property
    def object(self) -> Pattern:
        return Pattern(agents=self.parsed_agents)


@dataclass
class RuleBuilder(Visitor):
    """Builds Rule objects from Lark parse trees.

    Attributes:
        parsed_label: Optional label for the rule.
        left_agents: Agents on the left side of the rule.
        right_agents: Agents on the right side of the rule.
        parsed_rates: Rate expressions for the rule.
        tree_data: Type of rule being built.
    """

    parsed_label: Optional[str]
    left_agents: list[Optional[Agent]]
    right_agents: list[Optional[Agent]]
    parsed_rates: list[Expression]
    tree_data: str

    def __init__(self, tree: ParseTree):
        super().__init__()

        self.parsed_label = None
        self.left_agents = []
        self.right_agents = []
        self.parsed_rates = []

        assert tree.data in ["f_rule", "fr_rule", "ambi_rule", "ambi_fr_rule"]
        self.tree_data = tree.data

        self.visit(tree)

    # Visitor method for Lark
    def rate(self, tree: ParseTree) -> None:
        assert tree.data == "rate"

        expr = tree.children[0]
        assert expr.data == "algebraic_expression"

        rate = parse_tree_to_expression(expr)
        self.parsed_rates.append(rate)

    # Visitor method for Lark
    def rule_expression(self, tree: ParseTree) -> None:
        assert tree.data in ["rule_expression", "rev_rule_expression"]
        mid_idx = next(
            (i for i, child in enumerate(tree.children) if child in ["->", "<->"])
        )  # Locate the arrow in the expression

        for i, child in enumerate(tree.children):
            if i == mid_idx:
                continue

            if child == ".":
                agent = None
            elif child.data == "agent":
                agent = AgentBuilder(child).object

            if i < mid_idx:
                self.left_agents.append(agent)
            else:
                self.right_agents.append(agent)

    # Visitor method for Lark
    def rev_rule_expression(self, tree: ParseTree) -> None:
        self.rule_expression(tree)

    @property
    def objects(self) -> list[Rule]:
        rules = []
        left = Pattern(self.left_agents)
        right = Pattern(self.right_agents)
        rates = self.parsed_rates

        match self.tree_data:
            case "f_rule":
                assert len(rates) == 1
                rules.append(KappaRule(left, right, rates[0]))
            case "fr_rule":
                assert len(rates) == 2
                rules.append(KappaRule(left, right, rates[0]))
                rules.append(KappaRule(right, left, rates[1]))
            case "ambi_rule":
                # TODO: check that the order of the rates is right
                assert len(rates) == 2
                try:
                    assert rates[0].evaluate() == 0
                except:
                    rules.append(KappaRuleBimolecular(left, right, rates[0]))
                try:
                    assert rates[1].evaluate() == 0
                except:
                    rules.append(KappaRuleUnimolecular(left, right, rates[1]))
            case "ambi_fr_rule":
                assert len(rates) == 3
                try:
                    assert rates[0].evaluate() == 0
                except:
                    rules.append(KappaRuleBimolecular(left, right, rates[0]))
                try:
                    assert rates[1].evaluate() == 0
                except:
                    rules.append(KappaRuleUnimolecular(left, right, rates[1]))
                rules.append(KappaRule(right, left, rates[2]))

        return [r for r in rules if r is not None]


class LarkTreetoExpression(Transformer_NonRecursive):
    """Transforms a Lark ParseTree into an Expression object.

    Note:
        Uses a Transformer to preserve the tree structure of the original
        ParseTree. This doesn't need to use Transformer_NonRecursive anymore
        due to grammar changes, but methods explicitly call transform on children.
    """

    def algebraic_expression(self, children):
        children = [self.transform(c) for c in children]
        if len(children) == 1:
            return children[0]
        elif len(children) == 3 and children[0] == "(" and children[2] == ")":
            return children[1]
        else:
            raise Exception(f"Invalid algebraic expression: {children}")

    # --- Literals ---
    def SIGNED_FLOAT(self, token):
        return Expression("literal", value=float(token.value))

    def SIGNED_INT(self, token):
        return Expression("literal", value=int(token.value))

    # --- Variables/Constants ---
    def declared_variable_name(self, children):
        child = self.transform(children[0])
        return Expression("variable", name=child.value.strip("'\""))

    def reserved_variable_name(self, children):
        child = self.transform(children[0])
        return Expression("reserved_variable", value=child)

    def pattern(self, children):
        tree = Tree("pattern", children)
        pattern = PatternBuilder(tree).object
        assert (
            len(pattern.components) == 1
        ), "The pattern {pattern} must consist of a single component, since it is part of an Expression."
        component = pattern.components[0]

        return Expression("component_pattern", value=component)

    def defined_constant(self, children):
        child = self.transform(children[0])
        return Expression("defined_constant", name=child.value)

    # --- Operations ---
    def binary_op_expression(self, children):
        children = [self.transform(c) for c in children]
        left, op, right = children
        return Expression("binary_op", operator=op, left=left, right=right)

    def binary_op(self, children):
        return children[0]

    def unary_op_expression(self, children):
        children = [self.transform(c) for c in children]
        op, child = children
        return Expression("unary_op", operator=op, child=child)

    def unary_op(self, children):
        return children[0]

    def list_op_expression(self, children):
        children = [self.transform(c) for c in children]
        op_token, *args = children
        return Expression("list_op", operator=op_token.children[0], children=args)

    # --- Parentheses ---
    def parentheses(self, children):
        children = [self.transform(c) for c in children]
        return Expression("parentheses", child=children[0])

    # --- Ternary Conditional ---
    def conditional_expression(self, children):
        children = [self.transform(c) for c in children]
        cond, true_expr, false_expr = children
        cond = cond.children[0]
        return Expression(
            "conditional", condition=cond, true_expr=true_expr, false_expr=false_expr
        )

    # --- Boolean Logic ---
    def comparison(self, children):
        children = [self.transform(c) for c in children]
        left, op, right = children
        return Expression("comparison", operator=op.value, left=left, right=right)

    def logical_or(self, children):
        children = [self.transform(c) for c in children]
        left, right = children
        return Expression("logical_or", left=left, right=right)

    def logical_and(self, children):
        children = [self.transform(c) for c in children]
        left, right = children
        return Expression("logical_and", left=left, right=right)

    def logical_not(self, children):
        children = [self.transform(c) for c in children]
        return Expression("logical_not", child=children[0])

    # --- Boolean Literals ---
    def TRUE(self, token):
        return Expression("boolean_literal", value=True)

    def FALSE(self, token):
        return Expression("boolean_literal", value=False)

    # --- Default Fallthrough ---
    def __default__(self, data, children, meta):
        return Tree(data, children, meta)


def parse_tree_to_expression(tree: Tree) -> Expression:
    """Convert a Lark ParseTree to an Expression object.

    Note:
        Since there isn't extra logic when converting algebraic expressions,
        we can convert from the Lark representation in-place, without creating
        a new object, hence a Transformer instead of Visitor.

    Args:
        tree: Lark ParseTree rooted at algebraic_expression.

    Returns:
        Expression object representing the parsed expression.
    """
    return LarkTreetoExpression().transform(tree)



from dataclasses import dataclass, field
from typing import Optional, Iterable, Iterator, Self

from kappybara.pattern import Site, Agent, Component, Pattern, Embedding
from kappybara.utils import SetProperty, Property, IndexedSet


@dataclass(frozen=True)
class Edge:
    """Represents bonds between sites.

    Note:
        Edge(x, y) is the same as Edge(y, x).

    Attributes:
        site1: First site in the bond.
        site2: Second site in the bond.
    """

    site1: Site
    site2: Site

    def __eq__(self, other):
        return (self.site1 == other.site1 and self.site2 == other.site2) or (
            self.site1 == other.site2 and self.site2 == other.site1
        )

    def __hash__(self):
        return hash(frozenset((self.site1, self.site2)))


@dataclass
class Mixture:
    """A collection of agents and their connections.

    Attributes:
        agents: Indexed set of all agents in the mixture.
        _embeddings: Cache of embeddings for tracked components.
        _max_embedding_width: Maximum diameter of tracked components.
    """

    agents: IndexedSet[Agent]
    _embeddings: dict[Component, IndexedSet[Embedding]]
    _max_embedding_width: int

    @classmethod
    def from_kappa(cls, patterns: dict[str, int]) -> Self:
        """Create a mixture from Kappa pattern strings and counts.

        Args:
            patterns: Dictionary mapping pattern strings to copy counts.

        Returns:
            New Mixture with instantiated patterns.
        """
        real_patterns = []
        for pattern, count in patterns.items():
            real_patterns.extend([Pattern.from_kappa(pattern)] * count)
        return cls(real_patterns)

    def __init__(self, patterns: Optional[Iterable[Pattern]] = None):
        """Initialize a new mixture.

        Args:
            patterns: Optional collection of patterns to instantiate.
        """
        self.agents = IndexedSet()
        self._embeddings = {}
        self._max_embedding_width = 0

        self.agents.create_index("type", Property(lambda a: a.type))

        if patterns is not None:
            for pattern in patterns:
                self.instantiate(pattern)

    def __iter__(self) -> Iterator[Component]:
        yield from ComponentMixture([Pattern(list(self.agents))])

    @property
    def kappa_str(self) -> str:
        """The mixture representation in Kappa format.

        Returns:
            Kappa string with %init declarations for each component type.
        """
        return "\n".join(
            f"%init: {len(components)} {group.kappa_str}"
            for group, components in grouped(
                list(component for component in self)
            ).items()
        )

    def instantiate(self, pattern: Pattern | str, n_copies: int = 1) -> None:
        """Add instances of a pattern to the mixture.

        Args:
            pattern: Pattern to instantiate, or Kappa string.
            n_copies: Number of copies to create.

        Raises:
            AssertionError: If pattern is underspecified.
        """
        if isinstance(pattern, str):
            pattern = Pattern.from_kappa(pattern)

        assert (
            not pattern.underspecified
        ), "Pattern isn't specific enough to instantiate."
        for _ in range(n_copies):
            for component in pattern.components:
                self.add(component)

    def add(self, component: Component) -> None:
        """Add a component to the mixture.

        Args:
            component: Component to add with its agents and connections.
        """
        component_ordered = list(component.agents)
        new_agents = [agent.detached() for agent in component_ordered]
        new_edges = set()

        for i, agent in enumerate(component_ordered):
            # Duplicate the proper link structure
            for site in agent:
                if site.coupled:
                    partner = site.partner
                    i_partner = component_ordered.index(partner.agent)
                    new_site = new_agents[i][site.label]
                    new_partner = new_agents[i_partner][partner.label]
                    new_edges.add(Edge(new_site, new_partner))

        update = MixtureUpdate(agents_to_add=new_agents, edges_to_add=new_edges)
        self.apply_update(update)

    def remove(self, component: Component) -> None:
        """Remove a component from the mixture.

        Args:
            component: Component to remove.
        """
        update = MixtureUpdate()
        for agent in component:
            update.remove_agent(agent)
        self.apply_update(update)

    def embeddings(self, component: Component) -> IndexedSet[Embedding]:
        """Get embeddings of a tracked component.

        Args:
            component: Component to get embeddings for.

        Returns:
            Set of embeddings for the component.

        Raises:
            KeyError: If component is not being tracked.
        """
        try:
            return self._embeddings[component]
        except KeyError as e:
            e.add_note(
                f"Undeclared component: {component}. To embed it, first use `track_component`."
            )
            raise

    def track_component(self, component: Component):
        """Start tracking embeddings of a component.

        Args:
            component: Component pattern to track.
        """
        self._max_embedding_width = max(component.diameter, self._max_embedding_width)
        embeddings = IndexedSet(component.embeddings(self))
        embeddings.create_index("agent", SetProperty(lambda e: iter(e.values())))
        self._embeddings[component] = embeddings

    def apply_update(self, update: "MixtureUpdate") -> None:
        """Apply a collection of changes to the mixture.

        Args:
            update: MixtureUpdate specifying changes to apply.
        """
        for agent in update.touched_before:
            for tracked in self._embeddings:
                self._embeddings[tracked].remove_by("agent", agent)

        for edge in update.edges_to_remove:
            self._remove_edge(edge)
        for agent in update.agents_to_remove:
            self._remove_agent(agent)
        for agent in update.agents_to_add:
            self._add_agent(agent)
        for edge in update.edges_to_add:
            self._add_edge(edge)
        # NOTE: the current implementation doesn't directly mutate agent type

        update_region = neighborhood(update.touched_after, self._max_embedding_width)

        update_region = IndexedSet(update_region)
        update_region.create_index("type", Property(lambda a: a.type))
        for component_pattern in self._embeddings:
            new_embeddings = component_pattern.embeddings(update_region)
            for e in new_embeddings:
                self._embeddings[component_pattern].add(e)

    def _update_embeddings(self) -> None:
        for component_pattern in self._embeddings:
            self.track_component(component_pattern)

    def _add_agent(self, agent: Agent) -> None:
        """Add an agent to the mixture.

        Note:
            Calling these private functions isn't guaranteed to keep indexes
            up to date, which is why they shouldn't be used externally.
            The provided agent should not have any bound sites.

        Args:
            agent: Agent to add (should have empty sites).

        Raises:
            AssertionError: If agent has bound sites or isn't instantiable.
        """
        assert all(site.partner == "." for site in agent)  # Check all sites are unbound
        assert agent.instantiable
        self.agents.add(agent)

    def _remove_agent(self, agent: Agent) -> None:
        """Remove an agent from the mixture.

        Note:
            Any bonds associated with agent must be removed first.

        Args:
            agent: Agent to remove (should have empty sites).

        Raises:
            AssertionError: If agent has bound sites.
        """
        assert all(site.partner == "." for site in agent)  # Check all sites are unbound
        self.agents.remove(agent)

    def _add_edge(self, edge: Edge) -> None:
        """Add a bond between two sites.

        Args:
            edge: Edge specifying the bond to create.

        Raises:
            AssertionError: If either agent is not in the mixture.
        """
        assert edge.site1.agent in self.agents
        assert edge.site2.agent in self.agents
        edge.site1.partner = edge.site2
        edge.site2.partner = edge.site1

    def _remove_edge(self, edge: Edge) -> None:
        """Remove a bond between two sites.

        Args:
            edge: Edge specifying the bond to remove.

        Raises:
            AssertionError: If the edge doesn't exist.
        """
        assert edge.site1.partner == edge.site2
        assert edge.site2.partner == edge.site1
        edge.site1.partner = "."
        edge.site2.partner = "."


@dataclass
class ComponentMixture(Mixture):
    """A mixture that explicitly tracks connected components.

    Attributes:
        components: Indexed set of all components in the mixture.
    """

    components: IndexedSet[Component]

    def __init__(self, patterns: Optional[Iterable[Pattern]] = None):
        """Initialize a component-tracking mixture.

        Args:
            patterns: Optional collection of patterns to instantiate.
        """
        self.components = IndexedSet()
        self.components.create_index(
            "agent", SetProperty(lambda c: c.agents, is_unique=True)
        )
        super().__init__(patterns)

    def __iter__(self) -> Iterator[Component]:
        yield from self.components

    def embeddings_in_component(
        self, match_pattern: Component, mixture_component: Component
    ) -> list[dict[Agent, Agent]]:
        """Get embeddings of a pattern within a specific component.

        Args:
            match_pattern: Pattern to find embeddings for.
            mixture_component: Component to search within.

        Returns:
            List of embeddings within the specified component.
        """
        return self._embeddings[match_pattern].lookup("component", mixture_component)

    def track_component(self, component: Component):
        """Start tracking embeddings of a component pattern.

        Args:
            component: Component pattern to track.
        """
        super().track_component(component)
        self._embeddings[component].create_index(
            "component",
            Property(lambda e: self.components.lookup("agent", next(iter(e.values())))),
        )

    def apply_update(self, update: "MixtureUpdate") -> None:
        """Apply a collection of changes to the mixture.

        Args:
            update: MixtureUpdate specifying changes to apply.
        """
        super().apply_update(update)

    def _update_embeddings(self) -> None:
        for component_pattern in self._embeddings:
            self.track_component(component_pattern)

    def _add_agent(self, agent: Agent) -> None:
        """Add an agent as a new single-agent component.

        Args:
            agent: Agent to add.
        """
        super()._add_agent(agent)
        component = Component([agent])
        self.components.add(component)

    def _remove_agent(self, agent: Agent) -> None:
        """Remove an agent and its component.

        Args:
            agent: Agent to remove.

        Raises:
            AssertionError: If agent is part of a multi-agent component.
        """
        super()._remove_agent(agent)
        component = self.components.lookup("agent", agent)
        assert len(component) == 1
        self.components.remove(component)

    def _add_edge(self, edge: Edge) -> None:
        """Add an edge, potentially merging components.

        Args:
            edge: Edge to add between sites.
        """
        super()._add_edge(edge)

        # If the agents are in different components, merge the components
        # TODO: incremental mincut
        component1 = self.components.lookup("agent", edge.site1.agent)
        component2 = self.components.lookup("agent", edge.site2.agent)
        if component1 == component2:
            return

        # Ensure `component2` is the smaller of the 2
        if len(component2) > len(component1):
            component1, component2 = component2, component1

        relocated: dict[Component, list[Embedding]] = {}
        for tracked in self._embeddings:
            relocated[tracked] = list(
                self._embeddings[tracked].lookup("component", component2)
            )
            for e in relocated[tracked]:
                self._embeddings[tracked].remove(e)

        self.components.remove(component2)  # NOTE: invokes a redundant linear time pass
        for agent in component2:
            component1.add(agent)
            # TODO: better semantics for this type of operation
            #       Operate on diffs to set property.. ?
            self.components.indices["agent"][agent] = [component1]

        for tracked in self._embeddings:
            # TODO: refactor when we can register IndexedSet item updates, including
            # cached property evaluations
            for e in relocated[tracked]:
                assert (
                    self.components.lookup("agent", next(iter(e.values())))
                    == component1
                )
                self._embeddings[tracked].add(e)

    def _remove_edge(self, edge: Edge) -> None:
        """Remove an edge, potentially splitting components.

        Args:
            edge: Edge to remove.
        """
        super()._remove_edge(edge)

        agent1: Agent = edge.site1.agent
        agent2: Agent = edge.site2.agent
        old_component = self.components.lookup("agent", agent1)
        assert old_component == self.components.lookup("agent", agent2)

        # Create a new component if the old one got disconnected
        maybe_new_component = Component(agent1.depth_first_traversal)

        if agent2 in maybe_new_component:
            return  # The old component is still connected, do nothing

        new_component1 = maybe_new_component
        new_component2 = Component(agent2.depth_first_traversal)

        relocated: dict[Component, list[Embedding]] = {}
        for tracked in self._embeddings:
            relocated[tracked] = list(
                self._embeddings[tracked].lookup("component", old_component)
            )
            for e in relocated[tracked]:
                self._embeddings[tracked].remove(e)

        # TODO: need to do manual updates to the indices in `components`
        # to do this more efficiently
        self.components.remove(old_component)
        self.components.add(new_component1)
        self.components.add(new_component2)

        for tracked in self._embeddings:
            # TODO: refactor when we can register IndexedSet item updates, including
            # cached property evaluations
            for e in relocated[tracked]:
                assert self.components.lookup("agent", next(iter(e.values()))) in [
                    new_component1,
                    new_component2,
                ]
                self._embeddings[tracked].add(e)


@dataclass
class MixtureUpdate:
    """Specifies changes to be applied to a mixture.

    Attributes:
        agents_to_add: Agents to be added to the mixture.
        agents_to_remove: Agents to be removed from the mixture.
        edges_to_add: Edges to be created.
        edges_to_remove: Edges to be removed.
        agents_changed: Agents with internal state changes.
    """

    agents_to_add: list[Agent] = field(default_factory=list)
    agents_to_remove: list[Agent] = field(default_factory=list)
    edges_to_add: set[Edge] = field(default_factory=set)
    edges_to_remove: set[Edge] = field(default_factory=set)
    agents_changed: set[Agent] = field(default_factory=set)  # Agents changed internally

    def create_agent(self, agent: Agent) -> Agent:
        """Create a new agent based on a template.

        Note:
            Sites in the created agent will be emptied.

        Args:
            agent: Template agent to base the new agent on.

        Returns:
            New agent with empty sites.
        """
        new_agent = agent.detached()
        self.agents_to_add.append(new_agent)
        return new_agent

    def remove_agent(self, agent: Agent) -> None:
        """Specify to remove an agent and its edges from the mixture.

        Args:
            agent: Agent to remove.
        """
        self.agents_to_remove.append(agent)
        for site in agent:
            if site.coupled:
                self.edges_to_remove.add(Edge(site, site.partner))

    def connect_sites(self, site1: Site, site2: Site) -> None:
        """Specify to create an edge between two sites.

        If the sites are bound to other sites, indicates to remove those edges.

        Args:
            site1: First site to connect.
            site2: Second site to connect.
        """
        if site1.coupled and site1.partner != site2:
            self.disconnect_site(site1)
        if site2.coupled and site2.partner != site1:
            self.disconnect_site(site2)
        if not site1.partner == site2:
            self.edges_to_add.add(Edge(site1, site2))

    def disconnect_site(self, site: Site) -> None:
        """Specify that a site should be unbound.

        Args:
            site: Site to disconnect from its partner.
        """
        if site.coupled:
            self.edges_to_remove.add(Edge(site, site.partner))

    def register_changed_agent(self, agent: Agent) -> None:
        """Register an agent as having internal state changes.

        Args:
            agent: Agent that has been internally modified.
        """
        self.agents_changed.add(agent)

    @property
    def touched_after(self) -> set[Agent]:
        """The agents that will be changed or added after this update.

        Returns:
            Set of agents affected by the update.
        """
        touched = self.agents_changed | set(self.agents_to_add)

        for edge in self.edges_to_add:
            touched.add(edge.site1.agent)
            touched.add(edge.site2.agent)

        for edge in self.edges_to_remove:
            a, b = edge.site1.agent, edge.site2.agent
            if a not in self.agents_to_remove:  # TODO make agents_to_remove a set
                touched.add(a)
            if b not in self.agents_to_remove:
                touched.add(b)

        return touched

    @property
    def touched_before(self) -> set[Agent]:
        """The agents that will be changed or removed by this update.

        Returns:
            Set of agents affected before the update is applied.
        """
        touched = self.agents_changed | set(self.agents_to_remove)

        for edge in self.edges_to_remove:
            touched.add(edge.site1.agent)
            touched.add(edge.site2.agent)

        for edge in self.edges_to_add:
            a, b = edge.site1.agent, edge.site2.agent
            if a not in self.agents_to_add:  # TODO make agents_to_add a set
                touched.add(a)
            if b not in self.agents_to_add:
                touched.add(b)

        return touched


def neighborhood(agents: Iterable[Agent], radius: int) -> set[Agent]:
    """Get all agents within a distance radius of the given agents.

    Args:
        agents: Starting agents for the neighborhood.
        radius: Maximum distance to include.

    Returns:
        Set of all agents within the specified radius.
    """
    frontier = agents
    seen = set(frontier)
    for _ in range(radius):
        new_frontier = set()
        for cur in frontier:
            for n in cur.neighbors:
                seen.add(n)
                if n not in seen:
                    new_frontier.add(n)

        frontier = new_frontier
    return seen


def grouped(components: Iterable[Component]) -> dict[Component, list[Component]]:
    """Group components by isomorphism.

    Args:
        components: Components to group.

    Returns:
        Dictionary mapping representative components to lists of isomorphic components.
    """
    grouped: dict[Component, list[Component]] = {}
    for component in components:
        for group in grouped:
            if component.isomorphic(group):
                grouped[group].append(component)
                break
        else:
            grouped[component] = [component]
    return grouped



import pytest
from kappybara.system import System


@pytest.mark.parametrize(
    "expression, result",
    [
        ("[true] [?] 1 [:] 0", 1),
        ("[false] [?] 1 [:] 0", 0),
        ("[max] (1) (4)", 4),
        ("[min](1)(4)", 1),
    ],
)
def test_expression_evaluation(expression, result):
    assert System.from_ka(f"%obs: 'x' {expression}")["x"] == result



import pytest

from kappybara.pattern import Pattern
from kappybara.mixture import ComponentMixture


@pytest.mark.parametrize(
    "test_str",
    [
        "A(a[.]{u})",
        "A(a1[1]), B(b1[1])",
        "A(a1[1], a2[2], a3[5]), B(b1[2], b2[3]), C(c1[3], c2[4], c3[5]), D(d1[4], d2[1])",
        # "A(a1[1], a2[2], a3[5]), B(b1[2], b2[3])",
    ],
)
def test_instantiate_pattern_one_component(test_str):
    pattern = Pattern.from_kappa(test_str)
    mixture = ComponentMixture([pattern])
    assert pattern.components.pop().isomorphic(mixture.components.pop())


@pytest.mark.parametrize(
    "test_case",
    [
        ("A(a1[1]), B(b1[1])", 1, "A(a1[1]), B(b1[1])", 1),
        ("A(a1[1]), A(a1[1])", 1, "A(a1[1]), A(a1[1])", 2),
        ("A(a1[1]), A(a1[1]), A(a1[.]), A(a1[.])", 1, "A(a1[_])", 2),
        ("A(a1[1]), A(a1[1]), A(a1[.]), A(a1[.])", 1, "A(a1[.])", 2),
        ("A(a1[1]), A(a1[1]), A(a1[.]), A(a1[.])", 1, "A(a1[#])", 4),
        (
            "A(a1[1], a2[2], a3[5]), B(b1[2], b2[3]), C(c1[3], c2[4], c3[5]), D(d1[4], d2[1])",
            1,
            "A(a1[1], a2[2], a3[5]), B(b1[2], b2[3]), C(c1[3], c2[4], c3[5]), D(d1[4], d2[1])",
            1,
        ),
    ],
)
def test_find_embeddings_one_component(test_case):
    """
    Test embeddings of patterns consisting of a single component
    """
    n = 1000

    mixture_pattern_str, n_copies, match_pattern_str, n_embeddings_expected = test_case
    mixture_pattern = Pattern.from_kappa(mixture_pattern_str)

    mixture = ComponentMixture()
    for _ in range(n):
        mixture.instantiate(mixture_pattern, n_copies)

    match_pattern = Pattern.from_kappa(match_pattern_str)
    assert len(match_pattern.components) == 1

    embeddings = list(match_pattern.components[0].embeddings(mixture))
    assert len(embeddings) == n_embeddings_expected * n



import pytest
import math
from pathlib import Path

from kappybara.grammar import kappa_parser
from kappybara.pattern import Pattern
from kappybara.rule import KappaRule, KappaRuleUnimolecular, KappaRuleBimolecular
from kappybara.system import System

# Parser


def test_parse_file_without_error():
    kappa_parser.parse_file(
        str(Path(__file__).parent / "wnt_v8.ka")
    )  # 121 rules, 10 agents


@pytest.mark.parametrize(
    "test_kappa", ["A(s[.]), S(a[.]) -> A(s[1]), S(a[1])    @	1", "A(s[.])"]
)
def test_parse_without_error(test_kappa):
    kappa_parser.parse(test_kappa)


# Patterns


def test_pattern_from_kappa():
    test_kappa = """
        A(a[.]{blah}, b[_]{bleh}, c[#], d[some_site_name.some_agent_name], e[13]),
        B(f[13], e[1], z[3]),
        C(x[1]),
        D(w[3]),
        E()
    """
    pattern = Pattern.from_kappa(test_kappa)
    assert ["A", "B", "C", "D", "E"] == [agent.type for agent in pattern.agents]
    assert ["a", "b", "c", "d", "e"] == [site.label for site in pattern.agents[0]]
    assert len(pattern.components) == 2


# Rules


@pytest.mark.parametrize(
    "rule_str,rule_len",
    [
        ("A(a{p}), B(), . -> A(a{u}), B(), C() @ 1.0", 1),
        ("A(a{p}), B(), . <-> A(a{u}), B(), C() @ 1.0, 2.0", 2),
        ("A(b[.]), A(b[.]) <-> A(b[1]), A(b[1]) @ 100.0, 1.0", 2),
    ],
)
def test_rule_len_from_kappa(rule_str, rule_len):
    assert len(KappaRule.list_from_kappa(rule_str)) == rule_len


def test_ambi_rule_from_kappa():
    rules = KappaRule.list_from_kappa(
        "A(a{p}), B(b[1]), C(c[1]) -> A(a{u}), B(b[.]), C(c[.]) @ 1.0 {2.0}"
    )
    assert len(rules) == 2
    assert isinstance(rules[0], KappaRuleBimolecular)
    assert isinstance(rules[1], KappaRuleUnimolecular)
    assert rules[0].stochastic_rate.evaluate() == 1.0
    assert rules[1].stochastic_rate.evaluate() == 2.0


def test_uni_rule_from_kappa():
    rules = KappaRule.list_from_kappa(
        "A(a{p}), B(b[1]), C(c[1]) -> A(a{u}), B(b[.]), C(c[.]) @ 0.0 {2.0}"
    )
    assert len(rules) == 1
    assert isinstance(rules[0], KappaRuleUnimolecular)
    assert rules[0].stochastic_rate.evaluate() == 2.0


def test_bi_rule_from_kappa():
    rules = KappaRule.list_from_kappa(
        "A(a{p}), B(b[1]), C(c[1]) -> A(a{u}), B(b[.]), C(c[.]) @ 1.0 {0.0}"
    )
    assert len(rules) == 1
    assert isinstance(rules[0], KappaRuleBimolecular)
    assert rules[0].stochastic_rate.evaluate() == 1.0


def test_ambi_fr_rule_from_kappa():
    rules = KappaRule.list_from_kappa(
        "A(a{p}), B(b[1]), C(c[1]) <-> A(a{u}), B(b[.]), C(c[.]) @ 1.0 {2.0}, 3.0"
    )
    assert len(rules) == 3
    assert isinstance(rules[0], KappaRuleBimolecular)
    assert isinstance(rules[1], KappaRuleUnimolecular)
    assert rules[0].stochastic_rate.evaluate() == 1.0
    assert rules[1].stochastic_rate.evaluate() == 2.0
    assert rules[2].stochastic_rate.evaluate() == 3.0


# System


def test_system_kappa_str():
    system_in = System.from_ka(
        """
    %var: 'x'     0.03
    %var: 'k_on'  'x' * 10
    %var: 'g_on'  'k_on' / 100

    %var: 'n' 3 * 100
    %var: 'p' [pi] * 'n'
    %var: 'sqpi' [sqrt] ([pi])
    %var: 'm' [max] ('n') (5 * 3)

    %init: 'n' A(a[1]{p}), B(b[1]{u})

    %obs: 'pairs'   |A(a[1]), B(b[1])|

    A(a{p}), B(b[_]) -> A(a{u}), B() @ 'g_on'
        """
    )
    system_out = System.from_ka(system_in.kappa_str)
    for system in (system_in, system_out):
        assert system["pairs"] == system["n"] == system["m"] == 300
        assert system["p"] == pytest.approx(math.pi * system["n"])
        assert system["sqpi"] == pytest.approx(math.sqrt(math.pi))
    assert system_in["g_on"] == system_out["g_on"]



import os
import shutil
import tempfile
import random
import warnings
from collections import defaultdict
from functools import cached_property
from typing import Optional, Iterable, Self

import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.figure

from kappybara.mixture import Mixture, ComponentMixture
from kappybara.rule import Rule, KappaRule, KappaRuleUnimolecular, KappaRuleBimolecular
from kappybara.pattern import Component, Pattern
from kappybara.algebra import Expression


class System:
    """A Kappa system containing agents, rules, observables, and variables for simulation.

    Attributes:
        mixture: The current state of agents and their connections.
        rules: Dictionary mapping rule names to Rule objects.
        observables: Dictionary mapping observable names to expressions.
        variables: Dictionary mapping variable names to expressions.
        monitor: Optional Monitor object for tracking simulation history.
        time: Current simulation time.
        tallies: Dictionary tracking rule application counts.
    """

    mixture: Mixture
    rules: dict[str, Rule]
    observables: dict[str, Expression]
    variables: dict[str, Expression]
    monitor: Optional["Monitor"]
    time: float
    tallies: defaultdict[str, dict[str, int]]

    @classmethod
    def read_ka(cls, filepath: str) -> Self:
        """Read and parse a Kappa .ka file to create a System.

        Args:
            filepath: Path to the Kappa file.

        Returns:
            A new System instance parsed from the file.
        """
        with open(filepath) as f:
            return cls.from_ka(f.read())

    @classmethod
    def from_ka(cls, ka_str: str) -> Self:
        """Create a System from a Kappa (.ka style) string.

        Args:
            ka_str: Kappa language string containing a system definition.

        Returns:
            A new System instance parsed from the string.
        """
        from kappybara.grammar import (
            kappa_parser,
            parse_tree_to_expression,
            PatternBuilder,
            RuleBuilder,
        )

        input_tree = kappa_parser.parse(ka_str)
        assert input_tree.data == "kappa_input"

        variables: dict[str, Expression] = {}
        observables: dict[str, Expression] = {}
        rules: list[Rule] = []
        system_params: dict[str, int] = {}
        inits: list[tuple[Expression, Pattern]] = []

        for child in input_tree.children:
            tag = child.data

            if tag in ["f_rule", "fr_rule", "ambi_rule", "ambi_fr_rule"]:
                new_rules = RuleBuilder(child).objects
                rules.extend(new_rules)

            elif tag == "variable_declaration":
                name_tree = child.children[0]
                assert name_tree.data == "declared_variable_name"
                name = name_tree.children[0].value.strip("'\"")

                expr_tree = child.children[1]
                assert expr_tree.data == "algebraic_expression"
                value = parse_tree_to_expression(expr_tree)

                variables[name] = value

            elif tag == "plot_declaration":
                raise NotImplementedError

            elif tag == "observable_declaration":
                label_tree = child.children[0]
                assert isinstance(label_tree, str)
                name = label_tree.strip("'\"")

                expr_tree = child.children[1]
                assert expr_tree.data == "algebraic_expression"
                value = parse_tree_to_expression(expr_tree)

                observables[name] = value

            elif tag == "signature_declaration":
                raise NotImplementedError

            elif tag == "init_declaration":
                expr_tree = child.children[0]
                assert expr_tree.data == "algebraic_expression"
                amount = parse_tree_to_expression(expr_tree)

                pattern_tree = child.children[1]
                if pattern_tree.data == "declared_token_name":
                    raise NotImplementedError
                assert pattern_tree.data == "pattern"
                pattern = PatternBuilder(pattern_tree).object

                inits.append((amount, pattern))

            elif tag == "declared_token":
                raise NotImplementedError

            elif tag == "definition":
                reserved_name_tree = child.children[0]
                assert reserved_name_tree.data == "reserved_name"
                name = reserved_name_tree.children[0].value.strip("'\"")

                value_tree = child.children[1]
                assert value_tree.data == "value"
                value = int(value_tree.children[0].value)

                system_params[name] = value

            elif tag == "pattern":
                raise NotImplementedError

            else:
                raise TypeError(f"Unsupported input type: {tag}")

        system = cls(None, rules, observables, variables)
        for init in inits:
            system.mixture.instantiate(init[1], int(init[0].evaluate(system)))
        return system

    @classmethod
    def from_kappa(
        cls,
        mixture: Optional[dict[str, int]] = None,
        rules: Optional[Iterable[str]] = None,
        observables: Optional[list[str] | dict[str, str]] = None,
        variables: Optional[dict[str, str]] = None,
        *args,
        **kwargs,
    ) -> Self:
        """Create a System from Kappa strings.

        Args:
            mixture: Dictionary mapping agent patterns to initial counts.
            rules: Iterable of rule strings in Kappa format.
            observables: List of observable expressions or dict mapping names to expressions.
            variables: Dictionary mapping variable names to expressions.
            *args: Additional arguments passed to System constructor.
            **kwargs: Additional keyword arguments passed to System constructor.

        Returns:
            A new System instance.
        """
        real_rules = []
        if rules is not None:
            for rule in rules:
                real_rules.extend(KappaRule.list_from_kappa(rule))

        if observables is None:
            real_observables = {}
        elif isinstance(observables, list):
            real_observables = {
                f"o{i}": Expression.from_kappa(obs) for i, obs in enumerate(observables)
            }
        else:
            real_observables = {
                name: Expression.from_kappa(obs) for name, obs in observables.items()
            }

        real_variables = (
            {}
            if variables is None
            else {name: Expression.from_kappa(var) for name, var in variables.items()}
        )

        return cls(
            None if mixture is None else Mixture.from_kappa(mixture),
            real_rules,
            real_observables,
            real_variables,
            *args,
            **kwargs,
        )

    def __init__(
        self,
        mixture: Optional[Mixture] = None,
        rules: Optional[Iterable[Rule]] = None,
        observables: Optional[dict[str, Expression]] = None,
        variables: Optional[dict[str, Expression]] = None,
        monitor: bool = True,
    ):
        """Initialize a new System.

        Args:
            mixture: Initial mixture state.
            rules: Collection of rules to apply.
            observables: Dictionary of observable expressions.
            variables: Dictionary of variable expressions.
            monitor: Whether to enable monitoring of simulation history.
        """
        self.rules = (
            {} if rules is None else {f"r{i}": rule for i, rule in enumerate(rules)}
        )

        if not isinstance(mixture, ComponentMixture) and any(
            type(rule) in [KappaRuleUnimolecular, KappaRuleBimolecular]
            for rule in self.rules.values()
        ):
            patterns = [] if mixture is None else [Pattern(list(mixture.agents))]
            mixture = ComponentMixture(patterns)

        if mixture is None:
            mixture = (
                ComponentMixture()
                if any(
                    type(rule) in [KappaRuleUnimolecular, KappaRuleBimolecular]
                    for rule in self.rules
                )
                else Mixture()
            )

        self.observables = {} if observables is None else observables
        self.variables = {} if variables is None else variables

        self.set_mixture(mixture)
        self.time = 0

        self.tallies = defaultdict(lambda: {"applied": 0, "failed": 0})
        if monitor:
            self.monitor = Monitor(self)
            self.monitor.update()
        else:
            self.monitor = None

    def __str__(self):
        return self.kappa_str

    def __getitem__(self, name: str) -> int | float:
        """Get the value of an observable or variable.

        Args:
            name: Name of the observable or variable.

        Returns:
            Current value of the named expression.

        Raises:
            KeyError: If name doesn't correspond to any observable or variable.
        """
        if name in self.observables:
            return self.observables[name].evaluate(self)
        elif name in self.variables:
            return self.variables[name].evaluate(self)
        else:
            raise KeyError(
                f"Name {name} doesn't correspond to a declared observable or variable"
            )

    def __setitem__(self, name: str, kappa_str: str):
        """Set or update an observable or variable from a Kappa string.

        Args:
            name: Name to assign to the expression.
            kappa_str: Kappa expression string.
        """
        expr = Expression.from_kappa(kappa_str)
        self._track_expression(expr)
        if name in self.variables:
            self.variables[name] = expr
        else:  # Set new expressions as observables
            self.observables[name] = expr

    @property
    def names(self) -> dict[str, set[str]]:
        """The names of all observables and variables."""
        return {
            "observables": set(self.observables),
            "variables": set(self.variables),
        }

    @property
    def tallies_str(self) -> str:
        """A formatted string showing how many times each rule has been applied.

        Returns:
            Tab-separated string showing applied and failed counts for each rule.
        """
        return "Rule\tApplied\tFailed\n" + "\n".join(
            f"{rule_str}\t{tallies["applied"]}\t{tallies["failed"]}"
            for rule_str, tallies in self.tallies.items()
        )

    @property
    def kappa_str(self) -> str:
        """The system representation in Kappa (.ka style) format."""
        kappa_str = ""
        for var_name, var in self.variables.items():
            kappa_str += f"%var: '{var_name}' {var.kappa_str}\n"
        for rule in self.rules.values():
            assert isinstance(rule, KappaRule)
            kappa_str += f"{rule.kappa_str}\n"
        for obs_name, obs in self.observables.items():
            obs_str = (
                f"|{obs.kappa_str}|" if isinstance(obs, Component) else obs.kappa_str
            )
            kappa_str += f"%obs: '{obs_name}' {obs_str}\n"
        kappa_str += self.mixture.kappa_str
        return kappa_str

    def to_ka(self, filepath: str) -> None:
        """Write system information to a Kappa file.

        Args:
            filepath: Path where to write the Kappa file.
        """
        with open(filepath, "w") as f:
            f.write(self.kappa_str)

    def set_mixture(self, mixture: Mixture) -> None:
        """Set the system's mixture and update tracking.

        Args:
            mixture: New mixture to set for the system.
        """
        self.mixture = mixture
        for rule in self.rules.values():
            self._track_rule(rule)
        for observable in self.observables.values():
            self._track_expression(observable)
        for variable in self.variables.values():
            self._track_expression(variable)

    def add_rule(self, name: str, rule: Rule | str) -> None:
        """Add a new rule to the system.

        Args:
            name: Name to assign to the rule.
            rule: Rule object or Kappa string representation.

        Raises:
            AssertionError: If a rule with the given name already exists.
        """
        assert name not in self.rules, "Rule {name} already exists in the system"
        if isinstance(rule, str):
            rule = KappaRule.from_kappa(rule)
        self._track_rule(rule)
        self.rules[name] = rule

    def remove_rule(self, name: str) -> None:
        """Remove a rule by setting its rate to zero.

        Args:
            name: Name of the rule to remove.

        Raises:
            AssertionError: If the rule already has zero rate.
            KeyError: If no rule with the given name exists.
        """
        assert self.rules[name].rate(self) > 0, "Rule {name} is already null"
        try:
            self.rules[name].stochastic_rate = Expression.from_kappa("0")
        except KeyError as e:
            e.add_note("No rule {name} exists in the system")
            raise e

    def _track_rule(self, rule: Rule) -> None:
        """Track components mentioned in the left hand side of a Rule.

        Args:
            rule: Rule whose components should be tracked.
        """
        if isinstance(rule, KappaRule):
            for component in rule.left.components:
                # TODO: For efficiency check for isomorphism with already-tracked components
                self.mixture.track_component(component)

    def _track_expression(self, expression: Expression) -> None:
        """Track the Components in the given expression.

        Note:
            Doesn't track patterns nested by indirection - see the filter method.
        """
        for component_expr in expression.filter("component_pattern"):
            self.mixture.track_component(component_expr.attrs["value"])

    @cached_property
    def rule_reactivities(self) -> list[float]:
        """The reactivity of each rule in the system.

        Returns:
            List of reactivities corresponding to system rules.
        """
        return [rule.reactivity(self) for rule in self.rules.values()]

    @property
    def reactivity(self) -> float:
        """The total reactivity of the system.

        Returns:
            Sum of all rule reactivities.
        """
        return sum(self.rule_reactivities)

    def wait(self) -> None:
        """Advance simulation time according to exponential distribution.

        Raises:
            RuntimeWarning: If system has no reactivity (infinite wait time).
        """
        try:
            self.time += random.expovariate(self.reactivity)
        except ZeroDivisionError:
            warnings.warn(
                "system has no reactivity: infinite wait time", RuntimeWarning
            )

    def choose_rule(self) -> Optional[Rule]:
        """Choose a rule to apply based on reactivity weights.

        Returns:
            Selected rule, or None if no rules have positive reactivity.
        """
        try:
            return random.choices(
                list(self.rules.values()), weights=self.rule_reactivities
            )[0]
        except ValueError:
            warnings.warn("system has no reactivity: no rule applied", RuntimeWarning)
            return None

    def apply_rule(self, rule: Rule) -> None:
        """Apply a rule to the mixture and update tallies.

        Args:
            rule: Rule to apply to the current mixture.
        """
        update = rule.select(self.mixture)
        if update is not None:
            self.tallies[str(rule)]["applied"] += 1
            self.mixture.apply_update(update)
            del self.__dict__["rule_reactivities"]
        else:
            self.tallies[str(rule)]["failed"] += 1

    def update(self) -> None:
        """Perform one simulation step."""
        self.wait()
        if (rule := self.choose_rule()) is not None:
            self.apply_rule(rule)
        if self.monitor:
            self.monitor.update()

    def update_via_kasim(self, time: float) -> None:
        """Simulate for a given amount of time using KaSim.

        Note:
            KaSim must be installed and in the PATH.
            Some features may not be compatible between Kappybara and KaSim.

        Args:
            time: Additional time units to simulate.

        Raises:
            AssertionError: If KaSim is not found in PATH.
        """
        assert shutil.which(
            "KaSim"
        ), "To update via KaSim, it must be installed and in the PATH."

        with tempfile.TemporaryDirectory() as tmpdirname:
            # Run KaSim on the current system
            output_ka_path = os.path.join(tmpdirname, "out.ka")
            output_cmd = f'%mod: alarm {time} do $SNAPSHOT "{output_ka_path}";'
            input_ka_path = os.path.join(tmpdirname, "in.ka")
            with open(input_ka_path, "w") as f:
                f.write(f"{self.kappa_str}\n{output_cmd}")
            os.system(f"KaSim {input_ka_path} -l {time} -d {tmpdirname}")

            # Read the KaSim output
            output_kappa_str = ""
            with open(output_ka_path) as f:
                for line in f:
                    if line.startswith("%init"):
                        split = line.split("/")
                        output_kappa_str += split[0] + split[-1]

        # Apply the update
        self.set_mixture(System.from_ka(output_kappa_str).mixture)
        self.time += time
        if self.monitor:
            self.monitor.update()


class Monitor:
    """Records the history of the values of observables in a system.

    Attributes:
        system: The system being monitored.
        history: Dictionary mapping observable names to their value history.
    """

    system: System
    history: dict[str, list[Optional[float]]]

    def __init__(self, system: System):
        """Initialize a monitor for the given system.

        Args:
            system: System to monitor.
        """
        self.system = system
        self.history = {"time": []} | {obs_name: [] for obs_name in system.observables}

    def update(self) -> None:
        """Record current time and observable values."""
        self.history["time"].append(self.system.time)
        for obs_name in self.system.observables:
            if obs_name not in self.history:
                self.history[obs_name] = [None] * (len(self.history["time"]) - 1)
            self.history[obs_name].append(self.system[obs_name])

    def measure(self, observable_name: str, time: Optional[float] = None):
        """Get the value of an observable at a specific time.

        Args:
            observable_name: Name of the observable to measure.
            time: Time at which to measure. If None, uses latest time.

        Returns:
            Value of the observable at the specified time.

        Raises:
            AssertionError: If simulation hasn't reached the specified time.
        """
        times: list[int] = list(self.history["time"])
        if time is None:
            time = times[-1]
        assert time <= max(times), "Simulation hasn't reached time {time}"

        i = 0
        while times[i] < time:
            i += 1
        return self.history[observable_name][i]

    @property
    def dataframe(self) -> pd.DataFrame:
        """Get the history of observable values as a pandas DataFrame.

        Returns:
            DataFrame with time and observable columns.
        """
        return pd.DataFrame(self.history)

    def plot(self, combined: bool = False) -> matplotlib.figure.Figure:
        """Make a plot of all observables over time.

        Args:
            combined: Whether to plot all observables on the same axes.

        Returns:
            Matplotlib figure showing trajectories of observables.
        """
        if combined:
            fig, ax = plt.subplots()
            for obs_name in self.system.observables:
                ax.plot(self.history["time"], self.history[obs_name], label=obs_name)
            plt.legend()
            plt.xlabel("Time")
            plt.ylabel("Observable")
            plt.margins(0, 0)
        else:
            fig, axs = plt.subplots(len(self.system.observables), 1, sharex=True)
            for i, obs_name in enumerate(self.system.observables):
                axs[i].plot(self.history["time"], self.history[obs_name], color="black")
                axs[i].set_title(obs_name)
                if i == len(self.system.observables) - 1:
                    axs[i].set_xlabel("Time")
        return fig



import random

from kappybara.pattern import Component
from kappybara.system import System


def heterodimerization_system(k_on: float = 2.5e9) -> System:
    random.seed(42)
    avogadro = 6.0221413e23
    volume = 2.25e-12  # mammalian cell volume
    n_a, n_b = 1000, 1000
    return System.from_kappa(
        {"A(x[.])": n_a, "B(x[.])": n_b},
        rules=[
            f"A(x[.]), B(x[.]) -> A(x[1]), B(x[1]) @ {k_on / (avogadro * volume)}",
            "A(x[1]), B(x[1]) -> A(x[.]), B(x[.]) @ 2.5",
        ],
        observables=[f"|A(x[1]),B(x[1])|"],
    )



from collections import defaultdict
from functools import cached_property
from typing import Self, Optional, Iterator, Iterable, Union, NamedTuple, TYPE_CHECKING

from kappybara.utils import Counted, IndexedSet, Property

if TYPE_CHECKING:
    from kappybara.mixture import Mixture


# String partner states can be: "#" (wildcard), "." (empty), "_" (bound), "?" (undetermined)
# "?" is the default in pattern instantiation and a wildcard in rules and observations
SiteType = NamedTuple("SiteType", [("site_name", str), ("agent_name", str)])
Partner = str | SiteType | int | Union["Site"]


class Site(Counted):
    """Represents a site on an agent with state and binding partner information.

    Attributes:
        agent: The agent this site belongs to (set after initialization).
        label: Name of the site.
        state: Internal state of the site.
        partner: Binding partner specification.
    """

    agent: "Agent"  # Expected to be set after initialization

    def __init__(self, label: str, state: str, partner: Partner):
        """Initialize a site with label, state, and partner.

        Args:
            label: Name of the site.
            state: Internal state of the site.
            partner: Binding partner specification.
        """
        super().__init__()
        self.label = label
        self.state = state
        self.partner = partner

    def __repr__(self):
        return f'Site(id={self.id}, kappa_str="{self.kappa_str}")'

    @property
    def kappa_partner_str(self) -> str:
        if self.partner == "?":
            return ""
        elif self.coupled:
            return "[_]"
        return f"[{self.partner}]"

    @property
    def kappa_state_str(self) -> str:
        return "" if self.state == "?" else f"{{{self.state}}}"

    @property
    def kappa_str(self) -> str:
        """The site representation in Kappa format.

        Returns:
            Kappa string representation of the site.
        """
        return f"{self.label}{self.kappa_partner_str}{self.kappa_state_str}"

    @property
    def undetermined(self) -> bool:
        """Check if the site is in a state equivalent to leaving it unnamed in an agent.

        Returns:
            True if the site is undetermined.
        """
        return self.state == "?" and self.partner in ("?", ".")

    @property
    def underspecified(self) -> bool:
        """Check if a concrete Site can be created from this pattern.

        Returns:
            True if the site specification is incomplete.
        """
        return (
            self.state == "#"
            or self.partner in ("#", "_")
            or isinstance(self.partner, SiteType)
        )

    @property
    def stated(self) -> bool:
        """Check if the site has a specific internal state.

        Returns:
            True if the site has a determined state.
        """
        return self.state not in ("#", "?")

    @property
    def bound(self) -> bool:
        """Check if the site is bound.

        Returns:
            True if the site is bound to something.
        """
        return (
            self.partner == "_"
            or isinstance(self.partner, SiteType)
            or isinstance(self.partner, Site)
        )

    @property
    def coupled(self) -> bool:
        """Check if the site is coupled to a specific other site.

        Returns:
            True if the site has a direct partner reference.
        """
        return isinstance(self.partner, Site)

    def embeds_in(self, other: Self) -> bool:
        """Check whether self as a pattern matches other as a concrete site.

        Args:
            other: Concrete site to match against.

        Returns:
            True if this site pattern matches the concrete site.
        """
        if (self.stated and self.state != other.state) or (
            self.bound and not other.coupled
        ):
            return False

        match self.partner:
            case ".":
                return other.partner == "."
            case SiteType():
                return (
                    self.partner.site_name == other.partner.label
                    and self.partner.agent_name == other.partner.agent.type
                )
            case Site():
                return (
                    self.partner.agent.type == other.partner.agent.type
                    and self.label == other.label
                )

        return True


class Agent(Counted):
    """Represents an agent with a type and collection of sites.

    Attributes:
        type: Type name of the agent.
        interface: Dictionary mapping site labels to Site objects.
    """

    @classmethod
    def from_kappa(cls, kappa_str: str) -> Self:
        """Parse a single agent from a Kappa string.

        Args:
            kappa_str: Kappa string describing a single agent.

        Returns:
            Parsed Agent object.

        Raises:
            AssertionError: If the string doesn't describe exactly one agent.
        """
        from kappybara.grammar import kappa_parser, AgentBuilder

        # Check pattern describes only a single agent
        input_tree = kappa_parser.parse(kappa_str)
        assert input_tree.data == "kappa_input"
        assert len(input_tree.children) == 1
        pattern_tree = input_tree.children[0]
        assert pattern_tree.data == "pattern"
        assert (
            len(pattern_tree.children) == 1
        ), "Zero or more than one agent patterns were specified."
        agent_tree = pattern_tree.children[0]
        return AgentBuilder(agent_tree).object

    def __init__(self, type: str, sites: Iterable[Site]):
        """Initialize an agent with type and sites.

        Args:
            type: Type name of the agent.
            sites: Collection of sites belonging to this agent.
        """
        super().__init__()
        self.type = type
        self.interface = {site.label: site for site in sites}

    def __iter__(self):
        yield from self.sites

    def __getitem__(self, key: str) -> Site:
        """Get a site by its label.

        Args:
            key: Label of the site to retrieve.

        Returns:
            Site with the given label.
        """
        return self.interface[key]

    def __repr__(self):
        return f'Agent(id={self.id}, kappa_str="{self.kappa_str}")'

    @property
    def kappa_str(self):
        """The agent representation in Kappa format.

        Returns:
            Kappa string representation of the agent.
        """
        return f"{self.type}({" ".join(site.kappa_str for site in self)})"

    @property
    def sites(self) -> Iterable[Site]:
        """All sites of this agent.

        Yields:
            Sites belonging to this agent.
        """
        yield from self.interface.values()

    @cached_property
    def underspecified(self) -> bool:
        """Check if a concrete Agent can be created from this pattern.

        Returns:
            True if any site is underspecified.
        """
        return any(site.underspecified for site in self)

    @property
    def neighbors(self) -> list[Self]:
        """The agents directly connected to this one.

        Returns:
            List of neighboring agents.
        """
        return [site.partner.agent for site in self if site.coupled]

    @property
    def depth_first_traversal(self) -> list[Self]:
        """Perform depth-first traversal starting from this agent.

        Returns:
            List of agents in depth-first order.
        """
        visited = set()
        traversal = []
        stack = [self]
        while stack:
            if (agent := stack.pop()) not in visited:
                visited.add(agent)
                traversal.append(agent)
                stack.extend(agent.neighbors)
        return traversal

    @property
    def instantiable(self) -> bool:
        """Check if this agent pattern can be instantiated.

        Returns:
            True if all sites are sufficiently specified.
        """
        return not any(site.underspecified for site in self)

    def detached(self) -> Self:
        """Create a clone with all sites emptied of partners.

        Returns:
            New agent with same type and states but no connections.
        """
        detached = type(self)(
            self.type, [Site(site.label, site.state, ".") for site in self]
        )
        for site in detached:
            site.agent = detached
        return detached

    def isomorphic(self, other: Self) -> bool:
        """Check if two Agents are equivalent locally, ignoring partners.

        Note:
            Doesn't assume agents of the same type will have the same site signatures.

        Args:
            other: Agent to compare against.

        Returns:
            True if agents are locally isomorphic.
        """
        if self.type != other.type:
            return False

        b_sites_leftover = set(other.interface)
        for site_name, a_site in self.interface.items():
            # Check that `b` has a site with the same name and state
            if site_name not in other.interface and not a_site.undetermined:
                return False
            b_sites_leftover.remove(site_name)
            if a_site.state != other[site_name].state:
                return False

        # Check that sites in `other` not mentioned in `self`are undetermined
        return all(other[site_name].undetermined for site_name in b_sites_leftover)

    def embeds_in(self, other: Self) -> bool:
        """Check whether self as a pattern matches other as a concrete agent.

        Args:
            other: Concrete agent to match against.

        Returns:
            True if this agent pattern matches the concrete agent.
        """
        if self.type != other.type:
            return False

        for a_site in self:
            if a_site.label not in other.interface and not a_site.undetermined:
                return False
            b_site = other[a_site.label]
            if not a_site.embeds_in(b_site):
                return False

        return True


class Embedding(dict[Agent, Agent]):
    """Dictionary representing a mapping from pattern agents to mixture agents."""

    def __hash__(self):
        return hash(frozenset(self.items()))

    def __repr__(self):
        return f"Embedding({', '.join(f"{a.id}: {self[a].id}" for a in self)})"


class Component(Counted):
    """A set of agents that are all in the same connected component.

    Note:
        Connectedness is not guaranteed statically and must be enforced.

    Attributes:
        agents: Indexed set of agents in this component.
        n_copies: Number of copies of this component (usually 1).
    """

    agents: IndexedSet[Agent]
    n_copies: int

    @classmethod
    def from_kappa(cls, kappa_str: str) -> Self:
        """Parse a single component from a Kappa string.

        Args:
            kappa_str: Kappa string describing a connected component.

        Returns:
            Parsed Component object.

        Raises:
            AssertionError: If the pattern doesn't represent exactly one component.
        """
        parsed_pattern = Pattern.from_kappa(kappa_str)
        assert len(parsed_pattern.components) == 1
        return parsed_pattern.components[0]

    def __init__(self, agents: list[Agent], n_copies: int = 1):
        """Initialize a component with agents.

        Args:
            agents: List of agents in this component.
            n_copies: Number of copies of this component.

        Raises:
            AssertionError: If agents list is empty or n_copies < 1.
            NotImplementedError: If n_copies != 1 (not yet supported).
        """
        super().__init__()
        assert agents
        assert n_copies >= 1
        if n_copies != 1:
            raise NotImplementedError(
                "Simulations won't handle n_copies correctly in counting embeddings."
            )

        self.agents = IndexedSet(agents)  # TODO: order by graph traversal
        self.agents.create_index("type", Property(lambda a: a.type))
        self.n_copies = n_copies

    def __iter__(self):
        yield from self.agents

    def __len__(self):
        return len(self.agents)

    def __repr__(self):
        return f'Component(id={self.id}, kappa_str="{self.kappa_str}")'

    @property
    def kappa_str(self) -> str:
        """The component representation in Kappa format.

        Returns:
            Kappa string representation of the component.
        """
        return Pattern.agents_to_kappa_str(self.agents)

    def add(self, agent: Agent):
        """Add an agent to this component.

        Args:
            agent: Agent to add to the component.
        """
        self.agents.add(agent)

    def isomorphic(self, other: Self) -> bool:
        """Check if two components are isomorphic.

        Args:
            other: Component to compare against.

        Returns:
            True if an isomorphism exists between the components.
        """
        return next(self.isomorphisms(other), None) is not None

    def embeddings(
        self, other: Self | "Mixture" | Iterable[Agent], exact: bool = False
    ) -> Iterator[Embedding]:
        """Find embeddings of self in other.

        Args:
            other: Target to find embeddings in.
            exact: If True, finds isomorphisms instead of embeddings.

        Yields:
            Valid embeddings from self to other.
        """
        if hasattr(other, "agents"):
            other: IndexedSet[Agent] = other.agents

        assert "type" in other.properties

        a_root = next(iter(self.agents))  # "a" refers to `self` and "b" to `other`
        # Narrow the search by mapping `a_root` to agents in `other` of the same type
        for b_root in other.lookup("type", a_root.type):

            agent_map = Embedding({a_root: b_root})  # The potential bijection
            frontier = {a_root}
            root_failed = False

            while frontier and not root_failed:
                a = frontier.pop()
                b = agent_map[a]

                match_func = a.isomorphic if exact else a.embeds_in
                if not match_func(b):
                    root_failed = True
                    break

                for a_site in a:
                    b_site = b[a_site.label]

                    if a_site.coupled:
                        a_partner = a_site.partner.agent
                        b_partner = b_site.partner.agent

                        if b_partner not in other:
                            # The embedding must be enclosed within the set of agents
                            # provided.
                            root_failed = True
                            break
                        elif a_partner not in agent_map:
                            frontier.add(a_partner)
                            agent_map[a_partner] = b_partner
                        elif agent_map[a_site.partner.agent] != b_site.partner.agent:
                            root_failed = True
                            break
                    elif exact and a_site.partner != b_site.partner:
                        root_failed = True
                        break

            if not root_failed:
                yield agent_map  # A valid bijection

    def isomorphisms(self, other: Self | "Mixture") -> Iterator[dict[Agent, Agent]]:
        """Find bijections which respect links in the site graph.

        Checks for bijections ensuring that any internal site state specified
        in one component exists and is the same in the other.

        Note:
            Handles isomorphism generally, between instantiated components
            in a mixture and potentially between rule patterns.

        Args:
            other: Component or mixture to find isomorphisms with.

        Yields:
            Valid isomorphisms between the components.
        """
        if len(self.agents) != len(other.agents):
            return
        yield from self.embeddings(other, exact=True)

    @property
    def diameter(self) -> int:
        """Get the maximum minimum shortest path between any two agents.

        Returns:
            Diameter of the component graph.
        """

        def bfs_depth(root) -> int:
            frontier = set([root])
            seen = set()
            depth = -1

            while frontier:
                depth += 1
                new_frontier = set()
                seen = seen | frontier
                for cur in frontier:
                    for n in cur.neighbors:
                        if n not in seen:
                            new_frontier.add(n)

                frontier = new_frontier

            return depth

        return max(bfs_depth(a) for a in self.agents)


class Pattern:
    """A pattern consisting of multiple agents, some of which may be None (empty slots).

    Attributes:
        agents: List of agents, where None represents empty slots in rules.
    """

    agents: list[Optional[Agent]]

    @classmethod
    def from_kappa(cls, kappa_str: str) -> Self:
        """Parse a pattern from a Kappa string.

        Args:
            kappa_str: Kappa string describing a pattern.

        Returns:
            Parsed Pattern object.

        Raises:
            AssertionError: If the string doesn't describe exactly one pattern.
        """
        from kappybara.grammar import kappa_parser, PatternBuilder

        input_tree = kappa_parser.parse(kappa_str)
        assert input_tree.data == "kappa_input"
        assert (
            len(input_tree.children) == 1
        ), "Zero or more than one patterns were specified."
        assert len(input_tree.children) == 1
        pattern_tree = input_tree.children[0]
        return PatternBuilder(pattern_tree).object

    def __init__(self, agents: list[Optional[Agent]]):
        """Compile a pattern from a list of Agents.

        Replaces integer link states with references to actual partners, and
        constructs helper objects for tracking connected components. A None
        in agents represents an empty slot in a rule expression pattern.

        Args:
            agents: List of agents, where None represents empty slots.

        Raises:
            AssertionError: If integer links are malformed.
        """
        self.agents = agents

        # Parse site connections implied by integer LinkStates
        integer_links: defaultdict[int, list[Site]] = defaultdict(list)
        for agent in agents:
            if agent is not None:
                for site in agent:
                    if isinstance(site.partner, int):
                        integer_links[site.partner].append(site)

        # Replace integer LinkStates with Agent references
        for i in integer_links:
            linked_sites = integer_links[i]
            if len(linked_sites) == 1:
                raise AssertionError(f"Site link {i} is only referenced in one site.")
            elif len(linked_sites) > 2:
                raise AssertionError(
                    f"Site link {i} is referenced in more than two sites."
                )
            else:
                linked_sites[0].partner = linked_sites[1]
                linked_sites[1].partner = linked_sites[0]

    def __iter__(self) -> Iterator[Optional[Agent]]:
        yield from self.agents

    def __len__(self):
        return len(self.agents)

    @cached_property
    def components(self) -> list[Component]:
        """The connected components in this pattern.

        Returns:
            List of Component objects representing connected parts.
        """
        unseen = set(agent for agent in self.agents if agent is not None)
        components = []
        while unseen:
            component = Component(next(iter(unseen)).depth_first_traversal)
            unseen = unseen.difference(component)
            components.append(component)
        return components

    @staticmethod
    def agents_to_kappa_str(agents: Iterable[Optional[Agent]]) -> str:
        """Convert a collection of agents to Kappa string representation.

        Args:
            agents: Collection of agents to convert.

        Returns:
            Kappa string representation of the agents.
        """
        bond_num_counter = 1
        bond_nums: dict[Site, int] = dict()
        agent_strs = []
        for agent in agents:
            if agent is None:
                agent_strs.append(".")
                continue
            site_strs = []
            for site in agent:
                if site in bond_nums:
                    partner_str = f"[{bond_nums[site]}]"
                elif site.coupled:
                    partner_str = f"[{bond_num_counter}]"
                    bond_nums[site.partner] = bond_num_counter
                    bond_num_counter += 1
                else:
                    partner_str = "" if site.partner == "?" else f"[{site.partner}]"
                site_strs.append(f"{site.label}{partner_str}{site.kappa_state_str}")
            agent_strs.append(f"{agent.type}({" ".join(site_strs)})")
        return ", ".join(agent_strs)

    @property
    def kappa_str(self) -> str:
        """The pattern representation in Kappa format.

        Returns:
            Kappa string representation of the pattern.
        """
        return type(self).agents_to_kappa_str(self.agents)

    @cached_property
    def underspecified(self) -> bool:
        """Check if any agents in the pattern are underspecified.

        Returns:
            True if any agent is None or underspecified.
        """
        return any(agent is None or agent.underspecified for agent in self.agents)
